from django.shortcuts import render
from django.conf.urls import url
from django.views.decorators.csrf import csrf_exempt

from graphene_django.views import GraphQLView
from channels.routing import route_class
from graphql_ws.django_channels import GraphQLSubscriptionConsumer


def graphiql(request):
    graphiql_settings = {
        "GRAPHIQL_VERSION": "0.11.10",
        "SUBSCRIPTIONS_TRANSPORT_VERSION": "0.7.0",
        "subscriptionsEndpoint": "ws://localhost:8000/subscriptions",
        "endpointURL": "/graphql",
    }

    return render(request, "grapple/graphiql.html", graphiql_settings)


# Traditional URL routing
urlpatterns = [
    url(r"^graphql", csrf_exempt(GraphQLView.as_view(graphiql=True))),
    url(r"^graphiql", graphiql),
]

# Django Channel (v1.x) routing for subscription support
channel_routing = [route_class(GraphQLSubscriptionConsumer, path=r"^/subscriptions")]



import graphene
import wagtail
import wagtail.documents.blocks
import wagtail.embeds.blocks
import wagtail.images.blocks
import wagtail.snippets.blocks
from graphene.types import Scalar
from graphene_django.converter import convert_django_field
from wagtail.core.fields import StreamField
from wagtail.core import blocks

from ..registry import registry


class GenericStreamFieldInterface(Scalar):
    @staticmethod
    def serialize(stream_value):
        return stream_value.stream_data


@convert_django_field.register(StreamField)
def convert_stream_field(field, registry=None):
    return GenericStreamFieldInterface(description=field.help_text, required=not field.null)


class StreamFieldInterface(graphene.Interface):
    id = graphene.String()
    block_type = graphene.String()
    field = graphene.String()
    raw_value = graphene.String()

    @classmethod
    def resolve_type(cls, instance, info):
        """
        If block has a custom Graphene Node type in registry then use it,
        otherwise use generic block type.
        """
        mdl = type(instance.block)
        if mdl in registry.streamfield_blocks:
            return registry.streamfield_blocks[mdl]
        else:
            return registry.streamfield_blocks["generic-block"]

    def resolve_id(self, info, **kwargs):
        return self.id

    def resolve_block_type(self, info, **kwargs):
        return type(self.block).__name__

    def resolve_field(self, info, **kwargs):
        return self.block.name

    def resolve_raw_value(self, info, **kwargs):
        return self.value


def register_streamfield_blocks():
    from .pages import PageInterface
    from .documents import get_document_type
    from .images import get_image_type

    class StreamFieldBlock(graphene.ObjectType):
        value = graphene.String()

        class Meta:
            interfaces = (StreamFieldInterface,)

    class CharBlock(graphene.ObjectType):
        value = graphene.String()

        class Meta:
            interfaces = (StreamFieldInterface,)

    class TextBlock(graphene.ObjectType):
        value = graphene.String()

        class Meta:
            interfaces = (StreamFieldInterface,)

    class EmailBlock(graphene.ObjectType):
        value = graphene.String()

        class Meta:
            interfaces = (StreamFieldInterface,)

    class IntegerBlock(graphene.ObjectType):
        value = graphene.Int()

        class Meta:
            interfaces = (StreamFieldInterface,)

    class FloatBlock(graphene.ObjectType):
        value = graphene.Float()

        class Meta:
            interfaces = (StreamFieldInterface,)

    class DecimalBlock(graphene.ObjectType):
        value = graphene.Float()

        class Meta:
            interfaces = (StreamFieldInterface,)

    class RegexBlock(graphene.ObjectType):
        value = graphene.String()

        class Meta:
            interfaces = (StreamFieldInterface,)

    class URLBlock(graphene.ObjectType):
        value = graphene.String()

        class Meta:
            interfaces = (StreamFieldInterface,)

    class BooleanBlock(graphene.ObjectType):
        value = graphene.Boolean()

        class Meta:
            interfaces = (StreamFieldInterface,)

    class DateBlock(graphene.ObjectType):
        value = graphene.String(format=graphene.String())

        class Meta:
            interfaces = (StreamFieldInterface,)

        def resolve_value(self, info, **kwargs):
            format = kwargs.get("format")
            if format:
                return self.value.strftime(format)
            return self.value

    class DateTimeBlock(DateBlock):
        class Meta:
            interfaces = (StreamFieldInterface,)

    class TimeBlock(DateBlock):
        class Meta:
            interfaces = (StreamFieldInterface,)

    class RichTextBlock(graphene.ObjectType):
        value = graphene.String()

        class Meta:
            interfaces = (StreamFieldInterface,)

    class RawHTMLBlock(graphene.ObjectType):
        value = graphene.String()

        class Meta:
            interfaces = (StreamFieldInterface,)

    class BlockQuoteBlock(graphene.ObjectType):
        value = graphene.String()

        class Meta:
            interfaces = (StreamFieldInterface,)

    class ChoiceOption(graphene.ObjectType):
        key = graphene.String()
        value = graphene.String()

    class ChoiceBlock(graphene.ObjectType):
        value = graphene.String()
        choices = graphene.List(ChoiceOption)

        class Meta:
            interfaces = (StreamFieldInterface,)

        def resolve_choices(self, info, **kwargs):
            choices = []            
            for key, value in self.block._constructor_kwargs['choices']:
                choice = ChoiceOption(key, value)
                choices.append(choice)
            return choices

    class EmbedBlock(graphene.ObjectType):
        value = graphene.String()
        url = graphene.String()

        class Meta:
            interfaces = (StreamFieldInterface,)

        def resolve_url(self, info, **kwargs):
            return self.value.url

    class StaticBlock(graphene.ObjectType):
        value = graphene.String()

        class Meta:
            interfaces = (StreamFieldInterface,)

    class PageChooserBlock(graphene.ObjectType):
        page = graphene.Field(PageInterface)

        class Meta:
            interfaces = (StreamFieldInterface,)

        def resolve_page(self, info, **kwargs):
            return self.value

    class DocumentChooserBlock(graphene.ObjectType):
        document = graphene.Field(get_document_type())

        class Meta:
            interfaces = (StreamFieldInterface,)

        def resolve_document(self, info, **kwargs):
            return self.value

    class ImageChooserBlock(graphene.ObjectType):
        image = graphene.Field(get_image_type())

        class Meta:
            interfaces = (StreamFieldInterface,)

        def resolve_image(self, info, **kwargs):
            return self.value

    class SnippetChooserBlock(graphene.ObjectType):
        snippet = graphene.String()

        class Meta:
            interfaces = (StreamFieldInterface,)

        def resolve_snippet(self, info, **kwargs):
            return self.value

    class ListBlock(graphene.ObjectType):
        class Meta:
            interfaces = (StreamFieldInterface,)

        items = graphene.Field(StreamFieldInterface)

        def resolve_items(self, info, **kwargs):
            return self

    class StructBlockItem:
        id = None
        block = None

        def __init__(self, id, block):
            self.id = id
            self.block = block

    class StreamBlock(graphene.ObjectType):
        class Meta:
            interfaces = (StreamFieldInterface,)

        items = graphene.List(StreamFieldInterface)

        def resolve_items(self, info, **kwargs):
            return [
                StructBlockItem(name, block)
                for name, block in self.block.child_blocks.items()
            ]

    class StructBlock(graphene.ObjectType):
        class Meta:
            interfaces = (StreamFieldInterface,)

        items = graphene.List(StreamFieldInterface)

        def resolve_items(self, info, **kwargs):
            return [
                StructBlockItem(name, block)
                for name, block in self.block.child_blocks.items()
            ]

    registry.streamfield_blocks.update(
        {
            "generic-block": StreamFieldBlock,
            blocks.CharBlock: CharBlock,
            blocks.TextBlock: TextBlock,
            blocks.EmailBlock: EmailBlock,
            blocks.IntegerBlock: IntegerBlock,
            blocks.FloatBlock: FloatBlock,
            blocks.DecimalBlock: DecimalBlock,
            blocks.RegexBlock: RegexBlock,
            blocks.URLBlock: URLBlock,
            blocks.BooleanBlock: BooleanBlock,
            blocks.DateBlock: DateBlock,
            blocks.TimeBlock: TimeBlock,
            blocks.DateTimeBlock: DateTimeBlock,
            blocks.RichTextBlock: RichTextBlock,
            blocks.RawHTMLBlock: RawHTMLBlock,
            blocks.BlockQuoteBlock: BlockQuoteBlock,
            blocks.ChoiceBlock: ChoiceBlock,
            blocks.PageChooserBlock: PageChooserBlock,
            wagtail.documents.blocks.DocumentChooserBlock: DocumentChooserBlock,
            wagtail.images.blocks.ImageChooserBlock: ImageChooserBlock,
            wagtail.snippets.blocks.SnippetChooserBlock: SnippetChooserBlock,
            wagtail.embeds.blocks.EmbedBlock: EmbedBlock,
            blocks.StaticBlock: StaticBlock,
            blocks.ListBlock: ListBlock,
            blocks.StreamBlock: StreamBlock,
            blocks.StructBlock: StructBlock,
        }
    )



import graphene
import datetime
import urllib
import json

from django.db import models
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.core.signing import TimestampSigner
from django.shortcuts import render

from .registry import registry
from .signals import preview_update


# Classes used to define what the Django field should look like in the GQL type
class GraphQLField:
    field_name: str
    field_type: str

    def __init__(self, field_name: str, field_type: type = None):
        self.field_name = field_name
        if callable(field_type):
            self.field_type = field_type()
        else:
            self.field_type = field_type
        if field_type:
            self.field_type.source = field_name


def GraphQLString(field_name: str):
    class Mixin(GraphQLField):
        def __init__(self):
            super().__init__(field_name, graphene.String)

    return Mixin


def GraphQLFloat(field_name: str):
    class Mixin(GraphQLField):
        def __init__(self):
            super().__init__(field_name, graphene.Float)

    return Mixin


def GraphQLInt(field_name: str):
    class Mixin(GraphQLField):
        def __init__(self):
            super().__init__(field_name, graphene.Int)

    return Mixin


def GraphQLBoolean(field_name: str):
    class Mixin(GraphQLField):
        def __init__(self):
            super().__init__(field_name, graphene.Boolean)

    return Mixin


def GraphQLSnippet(field_name: str, snippet_model: str):
    class Mixin(GraphQLField):
        def __init__(self):
            (app_label, model) = snippet_model.lower().split(".")
            mdl = ContentType.objects.get(app_label=app_label, model=model)
            if mdl:
                self.field_type = registry.snippets[mdl.model_class()]
            else:
                self.field_type = graphene.String
            self.field_name = field_name

    return Mixin


def GraphQLStreamfield(field_name: str):
    from .types.streamfield import StreamFieldInterface
    
    class Mixin(GraphQLField):
        def __init__(self):
            super().__init__(field_name, graphene.List(StreamFieldInterface))

    return Mixin


def GraphQLImage(field_name: str):
    from wagtail.images import get_image_model_string
    
    return GraphQLForeignKey(field_name, get_image_model_string())


def GraphQLDocument(field_name: str):
    from django.conf import settings

    document_type = 'wagtaildocs.Document'
    if hasattr(settings, "WAGTAILDOCS_DOCUMENT_MODEL"):
        document_type = settings['WAGTAILDOCS_DOCUMENT_MODEL']

    return GraphQLForeignKey(field_name, document_type)


def GraphQLForeignKey(field_name: str, content_type: str, is_list: bool = False):
    class Mixin(GraphQLField):
        def __init__(self):
            app_label, model = content_type.lower().split(".")
            mdl = ContentType.objects.get(app_label=app_label, model=model)
            field_type = None

            if mdl:
                field_type = registry.models.get(mdl.model_class())

            if field_type and is_list:
                field_type = graphene.List(field_type)

            super().__init__(field_name, field_type)

    return Mixin


class PagePreview(models.Model):
    token = models.CharField(max_length=255, unique=True)
    content_type = models.ForeignKey(
        "contenttypes.ContentType", on_delete=models.CASCADE
    )
    content_json = models.TextField()
    created_at = models.DateField(auto_now_add=True)

    def as_page(self):
        content = json.loads(self.content_json)
        page_model = ContentType.objects.get_for_id(
            content["content_type"]
        ).model_class()
        page = page_model.from_json(self.content_json)
        page.pk = content["pk"]
        return page

    @classmethod
    def garbage_collect(cls):
        yesterday = datetime.datetime.now() - datetime.timedelta(hours=24)
        cls.objects.filter(created_at__lt=yesterday).delete()


# Mixin for pages that want extra Grapple benefits:
class GrapplePageMixin:
    @classmethod
    def get_preview_signer(cls):
        return TimestampSigner(salt="headlesspreview.token")

    def create_page_preview(self):
        if self.pk is None:
            identifier = "parent_id=%d;page_type=%s" % (
                self.get_parent().pk,
                self._meta.label,
            )
        else:
            identifier = "id=%d" % self.pk

        return PagePreview.objects.create(
            token=self.get_preview_signer().sign(identifier),
            content_type=self.content_type,
            content_json=self.to_json(),
        )

    def update_page_preview(self, token):
        return PagePreview.objects.update_or_create(
            token=token,
            defaults={
                "content_type": self.content_type,
                "content_json": self.to_json(),
            },
        )

    @classmethod
    def get_preview_url(cls, token):
        return f"{settings.GRAPPLE_PREVIEW_URL}?" + urllib.parse.urlencode(
            {
                "content_type": cls._meta.app_label + "." + cls.__name__.lower(),
                "token": token,
            }
        )

    def dummy_request(self, original_request=None, **meta):
        request = super(GrapplePageMixin, self).dummy_request(
            original_request=original_request, **meta
        )
        request.GET = request.GET.copy()
        request.GET["realtime_preview"] = original_request.GET.get("realtime_preview")
        return request

    def serve_preview(self, request, mode_name):
        token = request.COOKIES.get("used-token")
        is_realtime = request.GET.get("realtime_preview")

        if token and is_realtime:
            page_preview, existed = self.update_page_preview(token)
            PagePreview.garbage_collect()
            preview_update.send(sender=GrapplePageMixin, token=token)
        else:
            page_preview = self.create_page_preview()
            page_preview.save()
            PagePreview.garbage_collect()

        response = render(
            request,
            "grapple/preview.html",
            {"preview_url": self.get_preview_url(page_preview.token)},
        )
        
        # Set cookie that auto-expires after 5mins
        response.set_cookie(key="used-token", value=page_preview.token, max_age=300)
        return response

    @classmethod
    def get_page_from_preview_token(cls, token):
        content_type = ContentType.objects.get_for_model(cls)

        # Check token is valid
        cls.get_preview_signer().unsign(token)
        try:
            return PagePreview.objects.get(
                content_type=content_type, token=token
            ).as_page()
        except BaseException:
            return



import graphene
import inspect
from typing import Type
from django.db import models
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from wagtail.contrib.settings.models import BaseSetting
from wagtail.core.models import Page as WagtailPage
from wagtail.documents.models import AbstractDocument
from wagtail.images.models import AbstractImage
from wagtail.snippets.models import get_snippet_models
from graphene_django.types import DjangoObjectType


from .registry import registry
from .types.pages import PageInterface, Page
from .types.documents import DocumentObjectType
from .types.images import ImageObjectType


def import_apps():
    """
    Add each django app set in the settings file
    """
    apps = settings.GRAPPLE_APPS.items()
    for name, prefix in apps:
        add_app(name, prefix)
        registry.apps.append(name)


def add_app(app: str, prefix: str = ""):
    """
    Iterate through each model in the app and pass it to node type creators.
    """

    # Create a collection of models of standard models (Pages, Images,
    # Documents).
    models = [
        mdl.model_class() for mdl in ContentType.objects.filter(app_label=app).all()
    ]

    # Add snippet models to model collection.
    for snippet in get_snippet_models():
        if snippet._meta.app_label == app:
            models.append(snippet)

    # Create add each model to correct section of registry.
    for model in models:
        register_model(model, prefix)


def register_model(cls: type, type_prefix: str):
    """
    Pass model to the right node type creator based on it's base class.
    """

    # Pass class to correct type creator.
    if issubclass(cls, WagtailPage):
        register_page_model(cls, type_prefix)
    elif issubclass(cls, AbstractDocument):
        register_documment_model(cls, type_prefix)
    elif issubclass(cls, AbstractImage):
        register_image_model(cls, type_prefix)
    elif issubclass(cls, BaseSetting):
        register_settings_model(cls, type_prefix)
    elif cls in get_snippet_models():
        register_snippet_model(cls, type_prefix)
    else:
        register_django_model(cls, type_prefix)


def get_fields_and_properties(cls):
    """
    Return all fields and @property methods for a model.
    """
    fields = [field.name for field in cls._meta.get_fields(include_parents=False)]
    properties = []
    try:
        properties = [
            method[0]
            for method in inspect.getmembers(cls, lambda o: isinstance(o, property))
        ]
    except BaseException:
        properties = []

    return fields + properties


def build_node_type(
    cls: type,
    type_prefix: str,
    interface: graphene.Interface,
    base_type: Type[DjangoObjectType] = DjangoObjectType,
):
    """
    Build a graphene node type from a model class and associate
    with an interface. If it has custom fields then implmement them.
    """
    type_name = type_prefix + cls.__name__

    # Create a new blank node type
    class Meta:
        model = cls
        interfaces = (interface,) if interface is not None else tuple()
        exclude_fields = tuple()

    type_meta = {"Meta": Meta}
    type_meta.update({"id": graphene.ID()})

    # Build a list fields that shouldn't be reflected in GQL type.
    exclude_fields = get_fields_and_properties(cls)

    # Add any custom fields to node if they are defined.
    if hasattr(cls, "graphql_fields"):
        for field in cls.graphql_fields:
            if callable(field):
                field = field()

            # Take current field out of exclude_fields array
            if field.field_name in exclude_fields:
                exclude_fields.remove(field.field_name)

            # Add field to GQL type with correct field-type
            if field.field_type is not None:
                type_meta[field.field_name] = field.field_type

    # Set excluded fields to stop errors cropping up from unsupported field
    # types.
    type_meta["Meta"].exclude_fields = exclude_fields

    return type(type_name, (base_type,), type_meta)


def register_page_model(cls: Type[WagtailPage], type_prefix: str):
    """
    Create graphene node type for models than inherit from Wagtail Page model.
    """

    # Avoid gql type duplicates
    if cls in registry.pages:
        return

    # Create a GQL type derived from page model.
    page_node_type = build_node_type(cls, type_prefix, PageInterface, Page)

    # Add page type to registry.
    if page_node_type:
        registry.pages[cls] = page_node_type


def register_documment_model(cls: Type[AbstractDocument], type_prefix: str):
    """
    Create graphene node type for a model than inherits from AbstractDocument.
    Only one model will actually be generated because a default document model
    needs to be set in settings.
    """

    # Avoid gql type duplicates
    if cls in registry.documents:
        return

    # Create a GQL type derived from document model.
    document_node_type = build_node_type(cls, type_prefix, None, DocumentObjectType)

    # Add document type to registry.
    if document_node_type:
        registry.documents[cls] = document_node_type


def register_image_model(cls: Type[AbstractImage], type_prefix: str):
    """
    Create a graphene node type for a model than inherits from AbstractImage.
    Only one model will actually be generated because a default image model
    needs to be set in settings.
    """

    # Avoid gql type duplicates
    if cls in registry.images:
        return

    # Create a GQL type derived from document model.
    image_node_type = build_node_type(cls, type_prefix, None, ImageObjectType)

    # Add image type to registry.
    if image_node_type:
        registry.images[cls] = image_node_type


def register_settings_model(cls: Type[BaseSetting], type_prefix: str):
    """
    Create a graphene node type for a settings page.
    """

    # Avoid gql type duplicates
    if cls in registry.settings:
        return

    # Create a GQL type derived from document model.
    settings_node_type = build_node_type(cls, type_prefix, None)

    # Add image type to registry.
    if settings_node_type:
        registry.settings[cls] = settings_node_type


def register_snippet_model(cls: Type[models.Model], type_prefix: str):
    """
    Create a graphene type for a snippet model.
    """

    # Avoid gql type duplicates
    if cls in registry.snippets:
        return

    # Create a GQL type that implements Snippet Interface
    snippet_node_type = build_node_type(cls, type_prefix, None)

    if snippet_node_type:
        registry.snippets[cls] = snippet_node_type


def register_django_model(cls: Type[models.Model], type_prefix: str):
    """
    Create a graphene type for (non-specific) django model.
    Used for Orderables and other foreign keys.
    """

    # Avoid gql type duplicates
    if cls in registry.django_models:
        return

    # Create a GQL type that implements Snippet Interface
    django_node_type = build_node_type(cls, type_prefix, None)

    if django_node_type:
        registry.django_models[cls] = django_node_type



import graphene
from graphql.validation.rules import NoUnusedFragments, specified_rules

from .actions import import_apps
from .types.pages import PagesQuery, PagesSubscription
from .types.images import ImagesQuery
from .types.documents import DocumentsQuery
from .types.snippets import SnippetsQuery
from .types.settings import SettingsQuery
from .types.search import SearchQuery
from .types.streamfield import register_streamfield_blocks
from .registry import registry

"""
Import all the django apps defined in django settings then process each model
in these apps and create graphql node types from them.
"""
import_apps()
register_streamfield_blocks()


"""
Root schema object that graphene is pointed at.
It inherits its queries from each of the specific type mixins.
"""

# HACK: Remove NoUnusedFragments validator
# Due to the way previews work on the frontend, we need to pass all
# fragments into the query even if they're not used.
# This would usually cause a validation error. There doesn't appear
# to be a nice way to disable this validator so we monkey-patch it instead.


# We need to update specified_rules in-place so the change appears
# everywhere it's been imported

specified_rules[:] = [
    rule for rule in specified_rules
    if rule is not NoUnusedFragments
]

class Query(
    graphene.ObjectType,
    PagesQuery(),
    ImagesQuery(),
    DocumentsQuery(),
    SnippetsQuery(),
    SettingsQuery(),
    SearchQuery(),
):
    pass


class Subscription(PagesSubscription(), graphene.ObjectType):
    pass


schema = graphene.Schema(
    query=Query, types=list(registry.models.values()), subscription=Subscription
)



import graphene
from django.apps import apps
from wagtail.documents.models import get_document_model
from wagtail.images import get_image_model
from wagtail.search.backends import get_search_backend
from ..registry import registry


def SearchQuery():
    if registry.models:

        class Search(graphene.Union):
            class Meta:
                types = tuple(registry.models.values())

        class Mixin:
            search = graphene.List(Search, query=graphene.String())

            # Return just one setting base on name param.
            def resolve_search(self, info, **kwargs):
                query = kwargs.get('query')
                if query:
                    s = get_search_backend()
                    results = []
                    models = [get_document_model(), get_image_model()]
                    for app in registry.apps:
                        models += apps.all_models[app].values()
                    for model in models:
                        results += s.search(query, model)
                    return results
                return None

        return Mixin

    else:
        class Mixin:
            pass

        return Mixin



class RegistryItem(dict):
    @property
    def types(self) -> tuple:
        return tuple(self.values())


class Registry:
    apps = []
    pages = RegistryItem()
    documents = RegistryItem()
    images = RegistryItem()
    snippets = RegistryItem()
    streamfield_blocks = RegistryItem()
    django_models = RegistryItem()
    settings = RegistryItem()

    @property
    def models(self) -> dict:
        models: dict = {}
        models.update(self.pages)
        models.update(self.documents)
        models.update(self.images)
        models.update(self.snippets)
        models.update(self.streamfield_blocks)
        models.update(self.django_models)
        models.update(self.settings)
        return models


# Singleton Registry
registry = Registry()
